from datetime import datetime
from pathlib import Path
from typing import Optional

import pytest
from pydantic import ValidationError
from sqlalchemy.ext.asyncio import create_async_engine
from sqlmodel import Field, SQLModel, create_engine, or_, select
from sqlmodel.ext.asyncio.session import AsyncSession

from sadel import Sadel


class Hero(Sadel, table=True):
    __tablename__ = "hero"  # type: ignore
    _upsert_index_elements = {"id"}

    id: Optional[int] = Field(default=None, primary_key=True)
    name: str
    secret_name: str
    age: Optional[int] = None


@pytest.fixture
def temp_db(tmp_path: Path) -> Path:
    tmp_file = tmp_path / "database.db"
    temp_file_url = f"sqlite:///{tmp_file}"

    engine = create_engine(temp_file_url, echo=True, future=True)
    SQLModel.metadata.create_all(engine)
    return tmp_file


@pytest.mark.asyncio()
async def test_upsert(temp_db):
    sqlite_url_async = f"sqlite+aiosqlite:///{temp_db}"
    async_engine = create_async_engine(sqlite_url_async, echo=True, future=True)
    hero_1 = Hero(name="Deadpond", secret_name="Dive Wilson")
    async with AsyncSession(async_engine) as session:
        await Hero.upsert(hero_1, session)
        result = (await session.exec(select(Hero).where(Hero.name == "Deadpond"))).all()
        assert len(result) == 1
        assert result[0].name == "Deadpond"
        assert result[0].secret_name == "Dive Wilson"
        assert result[0].age is None
        assert result[0].id == 1
        assert isinstance(result[0].created_on, datetime)


@pytest.mark.asyncio()
async def test_batch_upsert(temp_db):
    sqlite_url_async = f"sqlite+aiosqlite:///{temp_db}"
    async_engine = create_async_engine(sqlite_url_async, echo=True, future=True)
    hero_1 = Hero(name="Deadpond", secret_name="Dive Wilson")
    hero_2 = Hero(name="Spider-Boy", secret_name="Pedro Parqueador")
    hero_3 = Hero(name="Rusty-Man", secret_name="Tommy Sharp", age=48)
    async with AsyncSession(async_engine) as session:
        await Hero.batch_upsert([hero_1, hero_2, hero_3], session)
        result = (
            await session.exec(
                select(Hero).where(
                    or_(
                        Hero.name == "Deadpond", Hero.name == "Spider-Boy", Hero.name == "Rusty-Man"
                    )
                )
            )
        ).all()
        assert len(result) == 3
        assert result[0].name == "Deadpond"
        assert result[1].name == "Spider-Boy"
        assert result[2].name == "Rusty-Man"


@pytest.mark.asyncio()
async def test_modified_on_upsert(temp_db):
    sqlite_url_async = f"sqlite+aiosqlite:///{temp_db}"
    async_engine = create_async_engine(sqlite_url_async, echo=True, future=True)
    hero_1 = Hero(id=1, name="Deadpond", secret_name="Dive Wilson")
    hero_1_update = Hero(id=1, name="Deadpond", secret_name="Dive Wilson", age=20)
    async with AsyncSession(async_engine) as session:
        await Hero.upsert(hero_1, session)
        await Hero.upsert(hero_1_update, session)
        [result] = (await session.exec(select(Hero).where(Hero.name == "Deadpond"))).all()

    assert isinstance(result.modified_on, datetime)
    assert result.age == 20


@pytest.mark.asyncio()
async def test_validate():
    with pytest.raises(ValidationError, match="Input should be a valid integer"):
        Hero(name="Deadpond", secret_name="Dive Wilson", age=datetime.now())



# pyright: reportUnknownVariableType=false
from __future__ import annotations

from datetime import datetime
from typing import Any, ClassVar

import sqlalchemy as sa
from pydantic import ConfigDict
from pydantic.v1 import validate_model
from sqlalchemy.dialects.postgresql import Insert, insert
from sqlmodel import Field, SQLModel


class Sadel(SQLModel):
    """Base class for SQL models."""

    model_config = ConfigDict(validate_assignment=True)  # type: ignore
    created_on: datetime | None = Field(
        default=None,
        sa_type=sa.DateTime(timezone=True),  # type: ignore
        sa_column_kwargs={"server_default": sa.func.now()},
        description=(
            "The date and time the record was created. "
            "Field is optional and not needed when instantiating a new record. "
            "It will be automatically set when the record is created in the database."
        ),
    )

    modified_on: datetime | None = Field(
        default=None,
        sa_type=sa.DateTime(timezone=True),  # pyright: ignore
        sa_column_kwargs={"onupdate": sa.func.now(), "server_default": sa.func.now()},
        description=(
            "The date and time the record was updated. "
            "Field is optional and not needed when instantiating a new record. "
            "It will be automatically set when the record is created in the database."
        ),
    )

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        # Workaround to validate the model on init, which is not supported by SQLModel
        if hasattr(self, "__config__"):
            _, _, validation_error = validate_model(self.__class__, data)  # pyright: ignore
            if validation_error:
                raise validation_error

    # Specifies the set of index elements which represent the ON CONFLICT target
    _upsert_index_elements: ClassVar[set[str]] = set()

    # Specifies the set of fields to exclude from updating in the resulting
    # UPSERT statement
    _upsert_exclude_fields: ClassVar[set[str]] = set()

    # Common fields which we should exclude when updating.
    _default_upsert_exclude_fields: ClassVar[set[str]] = {"created_on"}

    @classmethod
    async def upsert(cls, item: Sadel, session: sa.orm.Session):
        """upserts a single item"""
        stmt = cls._get_upsert_statement(item)
        await session.exec(stmt)
        await session.commit()

    @classmethod
    async def batch_upsert(cls, items: list[Sadel], session: sa.orm.Session):
        """Batch upserts a list of items."""
        for item in items:
            stmt = cls._get_upsert_statement(item)
            await session.exec(stmt)
        await session.commit()

    @classmethod
    def _get_upsert_statement(cls, item: Sadel) -> Insert:
        """Returns an UPSERT statement for a single item."""
        if not cls._upsert_index_elements:
            raise ValueError("No upsert index elements specified for the model.")

        to_insert = item.model_dump()
        to_insert["created_on"] = (
            sa.func.now()
        )  # set manually, because on_conflict_do_update doesn't trigger default oninsert
        to_update = cls._get_record_to_update(to_insert)
        stmt = insert(cls).values(to_insert)
        return stmt.on_conflict_do_update(
            index_elements=cls._upsert_index_elements,
            set_=to_update,
        )

    @classmethod
    def _get_record_to_update(cls, record: dict[str, Any]) -> dict[str, Any]:
        """Returns a record to be upserted taking into account the excluded fields."""
        exclude_fields = cls._upsert_exclude_fields.copy()
        exclude_fields.update(cls._default_upsert_exclude_fields)

        to_update = record.copy()

        for field in exclude_fields:
            _ = to_update.pop(field, None)  # pyright: ignore

        to_update["modified_on"] = (
            sa.func.now()
        )  # set manually. on_conflict_do_update doesn't trigger onupdate

        return to_update  # type: ignore
